#include "Okno.h"

//#include <gl\GL.h>
//#include <gl\GLU.h>
#include "glew.h"
#include "wglew.h"

#include "Rozszerzenia.h"

int WINAPI WinMain(HINSTANCE hInstance, 
				   HINSTANCE hPrevInstance,
				   LPSTR lpCmdLine, int nCmdShow)
{
	POINT polozenieOkna = {100,100};
	POINT rozmiarOkna = {800,600};
	if(!okno.Inicjuj(hInstance,polozenieOkna,rozmiarOkna,true,false))
	{
		MessageBox(NULL,"Inicjacja okna nie powioda si",
			"Aplikacja OpenGL", MB_OK | MB_ICONERROR);
		return EXIT_FAILURE;
	}
	else return okno.Uruchom();

	return 0;
}

LRESULT Okno::WndProc(HWND hWnd, UINT message,
					  WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_DESTROY: //zamykanie okna -> koczenie aplikacji
		 PostQuitMessage(EXIT_SUCCESS);
		 break;
	case WM_SIZE: //zmiana rozmiaru okna
		RECT rect;
		GetClientRect(hWnd, &rect);
		szerokoscObszaruUzytkownika = rect.right - rect.left;
		wysokoscObszaruUzytkownika = rect.bottom - rect.top;
		break;
	default: //automatyczne przetwarzanie komunikatw
		return DefWindowProc(hWnd, message, wParam, lParam);
	}

	return 0L;
}

bool Okno::Inicjuj(HINSTANCE uchwytAplikacji, 
				   POINT polozenieOkna, POINT rozmiarOkna, 
				   bool trybPelnoekranowy, bool zmianaRozdzielczosci)
{
	char nazwaOkna[] = "Aplikacja OpenGL";

	WNDCLASSEX wc;
	wc.cbSize=sizeof(wc);
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC; //styl okna
	wc.lpfnWndProc = (WNDPROC)::WndProc; //procedura okna
	wc.cbClsExtra = 0; //dodatkowe bajty zarezrwowane za klas okna
	wc.cbWndExtra = 0; //dodatkowe bajty zarezerwowane za instancj okna
	wc.hInstance = uchwytAplikacji; //instancja aplikacji
	wc.hIcon = NULL; //uchwyt ikony
	wc.hIconSm = NULL; //uchwyt ikony
	wc.hCursor = LoadCursor(NULL, IDC_ARROW); //uchwyt kursora
	wc.hbrBackground = NULL; //uchwyt pdzla ta
	wc.lpszMenuName = NULL; //nazwa menu
	wc.lpszClassName = nazwaOkna; //nazwa klasy okna

	//Rejestracja klasy okna
	if(RegisterClassEx(&wc)==0) return false;

	DWORD stylOkna = WS_OVERLAPPEDWINDOW;

	if(trybPelnoekranowy)
	{
		stylOkna = WS_POPUP;
		polozenieOkna.x = 0;
		polozenieOkna.y = 0;
		if(zmianaRozdzielczosci)
		{
			rozmiarOkna.x = 1024;
			rozmiarOkna.y = 768;
			if(!ZmianaRozdzielczosci(rozmiarOkna.x, rozmiarOkna.y)) return false;
		}
		else
		{
			RECT rozmiarEkranu;
			GetWindowRect(GetDesktopWindow(), &rozmiarEkranu);
			rozmiarOkna.x = rozmiarEkranu.right - rozmiarEkranu.left;
			rozmiarOkna.y = rozmiarEkranu.bottom - rozmiarEkranu.top;
		}
	}

	//Tworzenie okna
	uchwytOkna = CreateWindow(
		nazwaOkna, //nazwa klasy okna
		nazwaOkna, //nazwa okna
		stylOkna, //styl okna
		polozenieOkna.x, polozenieOkna.y, //pooenie okna (x,y)
		rozmiarOkna.x, rozmiarOkna.y, //rozmiar okna (szeroko, wysoko)
		NULL, //uchwyt okna nadrzdnego
		NULL, //uchwyt menu
		uchwytAplikacji, //uchwyt instancji aplikacji
		NULL); //parametr komunikatu WM_CREATE informujcego o utworzeniu okna

	if(uchwytOkna==NULL) return false;

	//Pokazanie i aktualizacja okna
	ShowWindow(uchwytOkna,SW_SHOW);
	UpdateWindow(uchwytOkna);

	return true;
}

bool Okno::ZmianaRozdzielczosci(long szerokosc, long wysokosc,
								long glebiaKolorow) const
{
	DEVMODE dmScreenSettings; //struktura trybu wywietlania 
	memset(&dmScreenSettings, 0, sizeof(dmScreenSettings)); //czyszczenie pamici
	dmScreenSettings.dmSize = sizeof(dmScreenSettings);
	dmScreenSettings.dmPelsWidth = szerokosc; //nowa szerokosc ekranu
	dmScreenSettings.dmPelsHeight = wysokosc; //nowa wysokosc ekranu
	dmScreenSettings.dmBitsPerPel = glebiaKolorow; //liczba bitw na piksel
	dmScreenSettings.dmFields = DM_BITSPERPEL |  DM_PELSWIDTH | DM_PELSHEIGHT;
	return ChangeDisplaySettings(&dmScreenSettings, CDS_FULLSCREEN) == DISP_CHANGE_SUCCESSFUL;
}

WPARAM Okno::Uruchom()
{
	//Ptla gwna - obsuga komunikatw
	MSG msg;
	while(GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}

//---------------------------------------

bool OknoGL::UstalFormatPikseli(HDC uchwytDC) const
{
	PIXELFORMATDESCRIPTOR opisFormatuPikseli;
	ZeroMemory(&opisFormatuPikseli, sizeof(opisFormatuPikseli));
	opisFormatuPikseli.nVersion = 1;
	opisFormatuPikseli.dwFlags = PFD_SUPPORT_OPENGL | 
		PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER; //w oknie, podwjne buforowanie
	opisFormatuPikseli.iPixelType = PFD_TYPE_RGBA; //tyo koloru RGB z kanaa alfa
	opisFormatuPikseli.cColorBits = 32; //jako kolorw, 4 bajty (po bajcie na kady kana)
	opisFormatuPikseli.cDepthBits = 32; //gboko bufora gbi (z-buffer)
	opisFormatuPikseli.iLayerType = PFD_MAIN_PLANE;
	int formatPikseli = ChoosePixelFormat(uchwytDC, &opisFormatuPikseli);
	if(formatPikseli == 0) return false;
	if(!SetPixelFormat(uchwytDC,formatPikseli,&opisFormatuPikseli)) return false;
	return true;
}

/*
bool OknoGL::InicjujWGL(HWND uchwytOkna)
{
	uchwytDC = ::GetDC(uchwytOkna);
	if (!UstalFormatPikseli(uchwytDC)) return false;

	uchwytRC = wglCreateContext(uchwytDC);
	if(uchwytRC == NULL) return false;
	if(!wglMakeCurrent(uchwytDC, uchwytRC)) return false;
	return true;
}
*/

bool OknoGL::InicjujWGL(HWND uchwytOkna)
{
	uchwytDC = ::GetDC(uchwytOkna);
	if (!UstalFormatPikseli(uchwytDC)) return false;

	HGLRC uchwytTymczasowegoRC = wglCreateContext(uchwytDC);
	if(uchwytTymczasowegoRC == NULL) return false;
	if(!wglMakeCurrent(uchwytDC, uchwytTymczasowegoRC)) return false;

	if(!InicjujRozszerzenia())
	{
		MessageBox(NULL, 
			"Pobranie adresw funkcji rozszerze ARB nie powiodo si", 
			"Aplikacja OpenGL",
			MB_OK | MB_ICONERROR);
		return false;
	}

	const int major_min = 3; const int minor_min = 3;
	int major, minor;
	glGetIntegerv(GL_MAJOR_VERSION, &major);
	glGetIntegerv(GL_MINOR_VERSION, &minor);
	if(major<major_min || (major==major_min && minor<minor_min))
	{
		MessageBox(NULL, 
			"Wersja OpenGL jest niewystarczajca", 
			"Aplikacja OpenGL",
			MB_OK | MB_ICONERROR);
		return false;
	}

	int atrybuty[] =
	{
		WGL_CONTEXT_MAJOR_VERSION_ARB, major_min,
		WGL_CONTEXT_MINOR_VERSION_ARB, minor_min,
		WGL_CONTEXT_FLAGS_ARB, WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB,
		WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB,
		//WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_CORE_PROFILE_BIT_ARB,
		0
	};

	uchwytRC = wglCreateContextAttribsARB(uchwytDC, 0, atrybuty);
	if(uchwytRC == NULL) return false;
	if(!wglMakeCurrent(uchwytDC, uchwytRC)) return false;

	wglDeleteContext(uchwytTymczasowegoRC);

	return true;
}


void OknoGL::UsuWGL()
{
	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(uchwytRC);
	::ReleaseDC(uchwytOkna, uchwytDC);
}

LRESULT OknoGL::WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	long wynik = Okno::WndProc(hWnd, message, wParam, lParam);

	switch(message)
	{
		case WM_CREATE: //Utworzenie okna
			//zmienna uchwytOkna nie jest jeszcze zainicjowana
			if(!InicjujWGL(hWnd))
			{
				MessageBox(NULL, "Pobranie kontekstu renderowania nie powiodo si", "Aplikacja OpenGL", MB_OK | MB_ICONERROR);
				return EXIT_FAILURE;
			}
			UmieInformacjeNaPaskuTytuu(hWnd);
			InicjujBuforWerteksw();
			UstawienieSceny();
			break;
		case WM_DESTROY: //Zamknicie okna
			UsuBuforWerteksw();
			UsuWGL();
			break;
		case WM_SIZE: //Zmiana rozmiaru okna
			UstawienieSceny();
			break;
		case WM_PAINT: //Okno wymaga odwieenia
			RysujScen();
			ValidateRect(hWnd, NULL);
			break;
	}

	return wynik;
}

void OknoGL::UmieInformacjeNaPaskuTytuu(HWND uchwytOkna)
{
	char bufor[256];
	GetWindowText(uchwytOkna, bufor, 256);
	const GLubyte* wersja = glGetString(GL_VERSION);
	strcat_s(bufor, " | OpenGL "); strcat_s(bufor, (char*)wersja);
	const GLubyte* dostawca = glGetString(GL_VENDOR);
	strcat_s(bufor, " | "); strcat_s(bufor, (char*)dostawca);
	const GLubyte* kartaGraficzna = glGetString(GL_RENDERER);
	strcat_s(bufor, " | "); strcat_s(bufor, (char*)kartaGraficzna);
	SetWindowText(uchwytOkna, bufor);
}

void OknoGL::UstawienieSceny(bool rzutowanieIzometryczne) //warto domylna = false
{
	glViewport(0,0, szerokoscObszaruUzytkownika, wysokoscObszaruUzytkownika); //okno OpenGL = wntrze formy (domylnie)

	//ustawienie puntu projekcji
	glMatrixMode(GL_PROJECTION); //przeczenie na macierz rzutowania
	glLoadIdentity();
	float wsp = wysokoscObszaruUzytkownika / (float)szerokoscObszaruUzytkownika;
	if(!rzutowanieIzometryczne)
		//left, right, bottom, top, znear, zfar (clipping)
		//mnoenie macierzy rzutowania przez macierz perspektywy - ustalanie frustum
		glFrustum(-1.0f, 1.0f, wsp*-1.0f, wsp*1.0f, 1.0f, 10.0f);
	else
		glOrtho(-1.0f, 1.0f, wsp*-1.0f, wsp*1.0f, 1.0f, 10.0f);
	glMatrixMode(GL_MODELVIEW);

	glEnable(GL_DEPTH_TEST); //z-buffer aktywny = ukrywanie przesonitych powierzchni
}

void OknoGL::RysujScen()
{
	/*
	const float x0 = 1.0f;
	const float y0 = 1.0f;
	const float z0 = 1.0f;
	*/

	//Przygotowanie buforw
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); //czyci bufory

	glLoadIdentity(); // macierz model-widok = macierz jednostkowa
	//glTranslatef(0.0f, 0.0f, -3.0f); //odsunicie kamery od sceny o 3
	/*
	glTranslatef(0.0f, 0.0f, -1.732f); //3
	glRotatef(35.26f, 1.0f, 0.0f, 0.0f); //2
	glRotatef(45.0f, 0.0f, 1.0f, 0.0f); //1
	*/
	/*
	gluLookAt(-1.0f, 1.0f, 1.0f,
		       0.0f, 0.0f, 0.0f,
			   0.0f, 1.0f, 0.0f);
	*/

	/*
	//ty kolor werteksw
	glColor4f(1.0f, 1.0f, 0.0f, 1.0f);

	//Rysowanie trjkta
	glBegin(GL_TRIANGLES);
	//ustalenie trzech wierzchokw trjkta (werteksw (x,y,z))
	//(0,0,z) jest mniej wicej w rodku ekranu
	//dolny lewy
	//glColor4f(1.0f, 1.0f, 0.0f, 1.0f);
	glVertex3f(-x0, -y0, 0.0f);
	//dolny prawy
	//glColor4f(1.0f, 0.0f, 1.0f, 1.0f);
	glVertex3f(x0, -y0, 0.0f); 
	//grny
	//glColor4f(0.0f, 1.0f, 1.0f, 1.0f);
	glVertex3f(0, y0, 0.0f); 
	//koniec rysowania figury
	glEnd();
	*/	

	//glDrawArrays(GL_TRIANGLES, 0, 3);
	
	//GLubyte indeksy[] = { 0, 1, 2 };
	//glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_BYTE, indeksy);

	GLubyte indeksy[] = { 0, 3, 1, 4 };
	glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_BYTE, indeksy);

	//Z bufora na ekran
	SwapBuffers(uchwytDC);
}

void OknoGL::InicjujBuforWerteksw()
{
	const float x0 = 1.0f;
	const float y0 = 1.0f;
	const float z0 = 1.0f;

	const float pooenia[] =
	{
		-x0, -y0, 0.0f, //dolny lewy
		 x0, -y0, 0.0f, //dolny prawy
		  0,  y0, 0.0f, //grny
		-x0,  y0, 0.0f, //grny lewy
		 x0,  y0, 0.0f  //grny prawy
	};

	const float kolory[] =
	{
		1, 1, 1, 1, //dolny lewy
		1, 0, 1, 1, //dolny prawy
		0, 1, 1, 1, //grny
		0, 1, 1, 1, //grny lewy
		1, 1, 1, 1  //grny prawy
	};

	//------------------

	//Vertex Array Object (VAO)
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	//Vertex Buffer Object (VBO)
	glGenBuffers(3, &vbo[0]); //tworzy bufory

	GLuint atrybutPooenia = 0;
	GLuint atrybutKolor = 3;

	glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(pooenia), pooenia, GL_STATIC_DRAW);
	glVertexAttribPointer(atrybutPooenia, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(atrybutPooenia);

	glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(kolory), kolory, GL_STATIC_DRAW);
	glVertexAttribPointer(atrybutKolor, 4, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(atrybutKolor);

	glBindBuffer(GL_ARRAY_BUFFER, NULL);
}

void OknoGL::UsuBuforWerteksw()
{
	glDeleteBuffers(3, vbo);
	glDeleteVertexArrays(1, &vao);
}